#include <stdio.h>
#include <stdlib.h>
#include <dir.h>

#define noDEBUG

typedef unsigned char      byte;
typedef unsigned short int word;
typedef unsigned long      ulong;

typedef struct
{
 int  c;
 int  h;
 int  r;
 int  n;
 int  f;
 int  Size;
 byte *Data;
} Msec __attribute__ ((packed));

#define MAXTRKS 32
typedef struct
{
 Msec S[2][MAXTRKS];
 int  N[2];       // Sector On Track
 int  dblfl[2];
 int  NormTbl[2][256];
} Mtrk __attribute__ ((packed));

int D_Track;
int D_Side;

Mtrk Dsk[90];

typedef struct
{
 byte TXT[2];
 byte SeqVal;
 byte ChkSig;
 byte TDVer;
 byte Dens;
 byte DrvType;
 byte TrkDens;
 byte DosMode;
 byte Surface;
 word CRC;
} HDR __attribute__ ((packed));

typedef struct
{
 byte SecPerTrk;
 byte PhisCyl;
 byte PhisSide;
 byte CRC;
} TRK __attribute__ ((packed));

typedef struct
{
 byte Cyl;
 byte Side;
 byte SNum;
 byte SLen;
 byte Syndrome;
 byte CRC;
} SEC __attribute__ ((packed));

typedef struct
{
 word CRC;
 word Len;
 byte YMD[3];
 byte HMS[3];
} CMT __attribute__ ((packed));

typedef struct
{
 char TXT[3];
 byte WR;
 word Cyl;
 word Head;
 word TxtOff;
 word DataOff;
 word Rezerv;
} FDIHdr __attribute__ ((packed));

typedef struct
{
 ulong TRKOff;
 word  Rez;
 byte  SPT;
} FDITrk __attribute__ ((packed));

typedef struct
{
 byte C;
 byte H;
 byte R;
 byte N;
 byte Flag;
 word SECOff;
} FDISec __attribute__ ((packed));


FDIHdr HFDI={"FDI",0,0,0,0,0,0};


byte BUF1[10240];
byte BUF2[102400];
HDR TDHdr;
FILE *InF,*OutF;

char ONAME[256]="";
#define FNAME argv[2]

#define ERR_Read -1
#define ERR_Struct -2
#define ERR_HDRCRC -3
#define ERR_TDPACK -4

void ReadComment(void)
{
 CMT C;
 int i,j;
 char *t;

 fread(&C,sizeof(CMT),1,InF);
 fread(BUF1,1,C.Len,InF);

 textattr(0x7);
 cprintf(" File creation data: ");
 textattr(0xf);
 cprintf("%02d.%02d.%d,  %d:%02d:%02d.\r\n",
        C.YMD[2],C.YMD[1],C.YMD[0]+1900,C.HMS[0],C.HMS[1],C.HMS[2]);

 textattr(0x7);

 t=BUF1;

 cprintf(" Comments:\r\n");
 textattr(0xf);
 cprintf("Ŀ\r\n");
// cprintf("+---------------------------------------------------------------------------+\r\n");


 textattr(0x7);
 for (i=0;i<8;i++)
    {
     if (j=strlen(t)) {cprintf("%s\r\n",t);t+=j;}
    }
 textattr(0xf);

 cprintf("\r\n");
 textattr(0xf);
// cprintf("---------------------------------------------------------------------------\r\n");


}

int ProcessHDR(void)
{
 if ( fread(&TDHdr,sizeof(HDR),1,InF) != 1) return ERR_Read;

 if ( CalcCRC((unsigned char *) &TDHdr,sizeof(HDR)-2) != TDHdr.CRC ) return ERR_HDRCRC;

 if ( strncmp(TDHdr.TXT,"td",2) == 0 ) return ERR_TDPACK;

 if ( ( strncmp(TDHdr.TXT,"TD",2) == 0 ) ) {

   textattr(0x7);
   cprintf(" TeleDisk version: ");
   textattr(0xf);
   cprintf("%d.%d\r\n",TDHdr.TDVer / 10,TDHdr.TDVer % 10);
   textattr(0x7);

   HFDI.Head=TDHdr.Surface;
   HFDI.Cyl=0;

   if ( (TDHdr.TrkDens & 0x80) == 0x80 ) ReadComment();

 } else return ERR_Struct;
 return 0;
}


int RLEExpand(byte *src,byte *dst,int slen)
{
 byte *s1,*s2;
 int len,len2,rlen;
 byte d1,d2;

 s2=dst;
 s1=src+3;

 switch (*(src+2))
   {
    case 0:
         {
           rlen=len=(BUF1[0] | BUF1[1] << 8)-1;
           while(len--) *s2++=*s1++;
           break;
         }
    case 1:
         {
           len=BUF1[3] | BUF1[4] << 8;
           rlen=len<<1;
           s1+=2;
           d1=*s1++;d2=*s1++;
           while (len--) {*s2++=d1;*s2++=d2;};
           break;
         }
    case 2:
         {
           len2=(BUF1[0] | BUF1[1] << 8)-1;
         l2:
            len2--;
            switch (*s1++)
              {
               case 0:
                    {
                     len=*s1++;len2--;
                     while (len--) {*s2++=*s1++;len2--;};
                     break;
                    }
               case 1:
                    {
                     len=*s1++;len2--;
                     d1=*s1++;d2=*s1++;len2-=2;
                     while (len--) {*s2++=d1;*s2++=d2;};
                     break;
                    }
              }
            if (len2 > 0) goto l2;
            rlen=s2-dst;
         }
   }
 return rlen;
}

int ReadFile(void)
{
 TRK T;
 SEC S;
 int i,j,len;
 word dlen;

 fread(&T,sizeof(TRK),1,InF);

 while ( (T.SecPerTrk != 255 ) && !feof(InF) )
   {
#ifdef DEBUG
    printf(" TRACK: PhisCyl:%2d, PhisSide:%2d, Sec:%2d\r\n",
           T.PhisCyl,T.PhisSide,T.SecPerTrk);
#endif

    if (T.PhisCyl>=HFDI.Cyl) HFDI.Cyl=T.PhisCyl;

    Dsk[T.PhisCyl].N[T.PhisSide]=T.SecPerTrk;

    for(i=0;i<T.SecPerTrk;i++)
       {

         fread(&S,sizeof(SEC),1,InF);
#ifdef DEBUG
         printf(" + Sector: C:%2d, H:%1d, R:%2d, Len:%d, Syn:%2d",
                  S.Cyl,S.Side,S.SNum,128<<S.SLen,S.Syndrome);
#endif

         Dsk[T.PhisCyl].S[T.PhisSide][i].c=S.Cyl;
         Dsk[T.PhisCyl].S[T.PhisSide][i].h=S.Side;
         Dsk[T.PhisCyl].S[T.PhisSide][i].r=S.SNum;
         Dsk[T.PhisCyl].S[T.PhisSide][i].n=S.SLen;

         if ( Dsk[T.PhisCyl].NormTbl[T.PhisSide][S.SNum] != -1 ) {
           Dsk[T.PhisCyl].dblfl[T.PhisSide]=1;
           }
         else Dsk[T.PhisCyl].NormTbl[T.PhisSide][S.SNum]=i;

         if ( ((S.Syndrome & 0x30) == 0) && ((S.SLen & 0xf8) == 0) ) // Sector Exist
           {
            fread(BUF1,1,2,InF);
            dlen=BUF1[0] | BUF1[1] << 8;
            if (dlen)
              {
               fread(BUF1+2,dlen,1,InF);
               len=RLEExpand(BUF1,BUF2,dlen);
               Dsk[T.PhisCyl].S[T.PhisSide][i].Size=len;
               Dsk[T.PhisCyl].S[T.PhisSide][i].Data=malloc(len);

               memmove(Dsk[T.PhisCyl].S[T.PhisSide][i].Data,BUF2,len);
#ifdef DEBUG
               printf(" pl=%d",len);
//               fwrite(BUF2,len,1,OutF);
#endif
              }
           }
#ifdef DEBUG
        printf("\n");
#endif
       }
    fread(&T,sizeof(TRK),1,InF);
   }

   HFDI.Cyl++; // bug, lost last track in fdi (14.07.99)
}


void ClearDSK(void)
{
 int t,s,h;

 for (t=0;t<90;t++)
    {
     Dsk[t].N[0]=-1;
     Dsk[t].N[1]=-1;
     Dsk[t].dblfl[0]=0;
     Dsk[t].dblfl[1]=0;

     for (h=0;h<2;h++)
     {
       for (s=0;s<MAXTRKS;s++)
          {
           Dsk[t].S[h][s].c=0xffffffff;
           Dsk[t].S[h][s].h=0xffffffff;
           Dsk[t].S[h][s].r=0xffffffff;
           Dsk[t].S[h][s].n=0xffffffff;
           Dsk[t].S[h][s].Size=0xffffffff;
           Dsk[t].S[h][s].Data=NULL;
         }
       for (s=0;s<256;s++)
          {
           Dsk[t].NormTbl[h][s]=-1;
          }
     }
    }
}

void FreeDSK(void)
{
 int t,s,h;

 for (t=0;t<90;t++)
  for (h=0;h<2;h++)
   for (s=0;s<MAXTRKS;s++)
    if (Dsk[t].S[h][s].Data !=NULL ) free(Dsk[t].S[h][s].Data);
}

int CheckTRD(void)
{
 int t,s,h;
 int TrdFl=1;


 for (t=80;t<89;t++)
     for (h=0;h<2;h++)
       if (Dsk[t].N[h] != -1 )
         {
          TrdFl=0;
          textattr(0xcf);
          cprintf("  Format error: ");
          textattr(0x7);
          cprintf(" data on track %d.\r\n",t);
          goto ret;
         }


 for (t=0;t<80;t++)
    {
     for (h=0;h<2;h++)
     {
       if (Dsk[t].N[h]==16)
         {
//           printf("\n %2d: n=%d, fl=%d: ",t,Dsk[t].N[h],Dsk[t].dblfl[h]);
           for (s=1;s<16;s++)
             {
              if ((Dsk[t].S[h][Dsk[t].NormTbl[h][s]].Size != 256) || (Dsk[t].S[h][Dsk[t].NormTbl[h][s]].r != s))
                {
                 TrdFl=0;
                 textattr(0xcf);
                 cprintf("  Format error: ");
                 textattr(0x7);
                 cprintf("missing or odd sector.\r\n");
                 goto ret;
                };
             }
         }
       else {
              TrdFl=0;
              textattr(0xcf);
              cprintf("  Format error: ");
              textattr(0x7);
              cprintf("track %d side %d have %d sector per track.\r\n",t,h,Dsk[t].N[h]);
              goto ret;
            };
     }
    }
ret:
 if (!TrdFl)
   {
    cprintf("  - sectors: ");
    for (s=0;s<Dsk[t].N[h];s++) cprintf("%d(%d) ", Dsk[t].S[h][s].r,Dsk[t].S[h][s].Size);
    cprintf("\r\n");
   }
 return TrdFl;
}

int MakeTRD(void)
{
 int t,s,h;
 int RetFl=1;
 int ss=0;

 for (t=0;t<80;t++)
  for (h=0;h<2;h++)
   for (s=1;s<17;s++)
   {
//    printf("%2d %2d %2d:%d\r\n ",t,h,s,Dsk[t].S[h][Dsk[t].NormTbl[h][s]].Size);
//    ss+=Dsk[t].S[h][Dsk[t].NormTbl[h][s]].Size;
      if (
          fwrite(Dsk[t].S[h][Dsk[t].NormTbl[h][s]].Data,
                 Dsk[t].S[h][Dsk[t].NormTbl[h][s]].Size,1,OutF) != 1
         ) RetFl=0;
   }
// printf("\r\n size=%d",ss);
 return RetFl;
}

int MakeFDI(void)
{
 int t,s,h;
 int RetFl=1;
 int ss=0;

 FDITrk TT;
 FDISec SS;

 ulong Off=0;
 ulong OffSect=0;

 char AddText[]="\r\n'This file created by TDCVT by Sergey Erokhin'\r\n";

 fwrite(&HFDI,sizeof(FDIHdr),1,OutF);

 for (t=0;t<HFDI.Cyl;t++)
  for (h=0;h<HFDI.Head;h++)
  {
   Off+=OffSect;

   TT.TRKOff=Off;
   TT.Rez=0;
   TT.SPT=Dsk[t].N[h];

   fwrite(&TT,sizeof(FDITrk),1,OutF);

   OffSect=0;

   if (TT.SPT)
    for (s=0;s<TT.SPT;s++)
    {
     SS.C=Dsk[t].S[h][s].c;
     SS.H=Dsk[t].S[h][s].h;
     SS.R=Dsk[t].S[h][s].r;
     SS.N=Dsk[t].S[h][s].n;
     SS.Flag=1<<SS.N;

     SS.SECOff=OffSect;

     OffSect+=128<<SS.N;
     fwrite(&SS,sizeof(FDISec),1,OutF);
    }
#ifdef DEBUG
     printf("Cyl:%02d Head:%d N:%d\n",t,h,TT.SPT);
#endif
  }


 HFDI.TxtOff=ftell(OutF);
 fwrite(AddText,sizeof(AddText),1,OutF);
 HFDI.DataOff=ftell(OutF);

 for (t=0;t<HFDI.Cyl;t++)
  for (h=0;h<HFDI.Head;h++)
   for (s=0;s<Dsk[t].N[h];s++)
    {
     if (Dsk[t].S[h][s].Data)
       fwrite(Dsk[t].S[h][s].Data,Dsk[t].S[h][s].Size,1,OutF);
    }

 rewind(OutF);
//  fseek(OutF,0,SEEK_SET);

 fwrite(&HFDI,sizeof(FDIHdr),1,OutF);

 return RetFl;
}


void MakeDstFNAME(char *src, char *dst,int TRD, int FDI)
{
 char d[MAXDRIVE], p[MAXDIR], f[MAXFILE], e[MAXEXT];
 int which = fnsplit(src, d, p, f, e);

 strcpy(e,".???");
 if (TRD) strcpy(e,".TRD");
 if (FDI) strcpy(e,".FDI");
 fnmerge(dst,d, p, f, e);
}

void Usage(void)
{
  cprintf(" usage: TDCVT -{Flag} InFile [OutFile]\r\n\r\n");
  cprintf("          Flag: -? - this help\r\n");
  cprintf("                -T - convert to .TRD\r\n");
  cprintf("                -F - convert to .FDI\r\n");

  exit(-1);
}

int main(int argc,char **argv)
{

 int TRDFlag=0;
 int FDIFlag=0;

 int DeleteOUT=1;

 int i;

 textattr(0x0F);
 cprintf("Ŀ\r\n");
 textattr(0x02);
 cprintf(" .TD? to .TRD,.FDI Convertor V0.11 by Sergey Erokhin, "__DATE__);
 textattr(0x0F);
 cprintf(" \r\n\r\n");
 textattr(0x07);


 if ( argc < 2 ) Usage();

 strupr(argv[1]);

 if (!strcmp(argv[1],"-T")) TRDFlag=1;
 if (!strcmp(argv[1],"-F")) FDIFlag=1;
 if ( (!TRDFlag && !FDIFlag) || (!strcmp(argv[1],"-?")) ) Usage();



 if ( (argc == 3) && (TRDFlag || FDIFlag) ) MakeDstFNAME(argv[2],ONAME,TRDFlag,FDIFlag);
   else if (argc == 4) strcpy(ONAME,argv[3]);
     else Usage();

 strupr(FNAME);
 strupr(ONAME);


 if ( (InF=fopen(FNAME,"rb")) == NULL ) {
   cprintf("Can't open file %s\r\n",FNAME);
   exit(-1);
 }

 cprintf(" %s -> %s\r\n",FNAME,ONAME);
 cprintf("\r\n");

 ClearDSK();

 if ( i=ProcessHDR() )
   {
    textattr(0xcf);
    cprintf(" ! Error:");
    textattr(0x7);
    switch (i)
      {
       case ERR_Read:   {cprintf(" Read file");break;}
       case ERR_Struct: {cprintf(" structure error.");break;}
       case ERR_HDRCRC: {cprintf(" header CRC.");break;}
       case ERR_TDPACK: {cprintf(" Packed file, please use TDUNP first.");break;}

      }
    cprintf("\r\n Bye.\r\n");
    exit(-1);
   }

 textattr(0xf);
 cprintf("\n Reading image in memory.\r\n");
 textattr(0x7);
 ReadFile();


 if ( (OutF=fopen(ONAME,"wb")) == NULL ) {
   cprintf("Can't open file %s\r\n",ONAME);
   exit(-1);
 }

 if ( TRDFlag )
 {
   textattr(0xf);
   cprintf("\n Convert .TD? in .TRD\r\n");
   textattr(0x7);
   if ( CheckTRD() )
     {
      if (!MakeTRD()) {textattr(0xcf);cprintf("Write error.\r\n");textattr(0x7);}
      else {textattr(0xf);cprintf(" File create ok.\r\n");DeleteOUT=0;};
     }
   else cprintf(" sorry, file complex for .TRD\r\n");
 }

 if ( FDIFlag )
 {
   textattr(0xf);
   cprintf("\n Convert .TD? to .FDI\r\n");
   textattr(0x7);
   MakeFDI();
   DeleteOUT=0;
/*   if ( CheckTRD() )
     {
      if (!MakeTRD()) cprintf("訡 .\n");
      else {
*/
   cprintf(" File create ok.\n");/*DeleteOUT=0;};
     }
*/
 }

 FreeDSK();

 fclose(InF);
 fclose(OutF);

 if (DeleteOUT) unlink(ONAME);

/*
 OutF=fopen("DUMP","wb");
 fwrite(Dsk,sizeof(Dsk),1,OutF);
 fclose(OutF);
*/
 textattr(0x07);cprintf("\r\n");

 return 0;
}